import ipaddress
import re
from collections.abc import Set
from operator import itemgetter

from bgpfu.base import BaseObject
from bgpfu.prefixlist import PrefixListBase


class PrefixSet(PrefixListBase, Set):
    def __init__(self, data=None, **kwargs):
        super().__init__()
        self.log_init()
        # if we weren't given a dict, try and parse it out
        # as prefix/len^min-max into the expected dict structure
        if data is None:
            data = {}
        if not isinstance(data, dict):
            try:
                data = self.parse_prefix_range(expr=data)
            except ValueError as e:
                self.log.error(msg=str(e))
                raise e
        # save meta information
        self._meta = kwargs
        self.log.debug(msg="creating prefix sets")
        # create sets per address family to hold prefix range tuples as
        # lower and upper bounds for prefix indices
        self._sets = {"ipv4": set(), "ipv6": set()}
        for af in data:
            # determine the ip address version that we're dealing with
            try:
                version = int(re.match(r"^ipv(4|6)", af).group(1))
            except (ValueError, AttributeError):
                self.log.warning(msg="invalid address-family %s" % af)
                continue
            self.log.debug(msg="adding %s prefixes to set" % af)
            # create a temporary list to hold index ranges
            temp = list()
            for entry in data[af]:
                # calculate the index of the base prefix
                # we'll be calulating a breadth-first index
                # of the CBT sub-tree rooted at this node
                try:
                    prefix, root = self.index_of(entry["prefix"])
                except ValueError as e:
                    self.log.warning(msg=str(e))
                    continue
                # check that the prefix has the correct address version
                if prefix.version != version:
                    self.log.warning(
                        msg="prefix %s not of version %d" % (prefix, version)
                    )
                    continue
                # get the length of the base prefix and the maximum prefix length
                # for the given address family
                l = prefix.prefixlen
                h = prefix.max_prefixlen
                self.log.debug(
                    msg="setting base index of prefix: %s = %d" % (prefix, root)
                )
                # check if we have been given min and max prefix-lengths
                m_set = False
                try:
                    m = max(int(entry["greater-equal"]), l)
                    m_set = True
                except (KeyError, ValueError):
                    m = l
                self.log.debug(msg="min-length set to %d" % m)
                try:
                    n = min(int(entry["less-equal"]), h)
                except (KeyError, ValueError):
                    if m_set:
                        n = h
                    else:
                        n = m
                self.log.debug(msg="max-length set to %d" % n)
                self.log.debug(msg="traversing subtree from index %d" % root)
                # calculate the total depth of the iteration
                depth = n - l
                left = right = root
                for d in range(0, depth + 1):
                    # check whether we're yet at the depth corresponding to the
                    # minimum prefix-length provided
                    if d >= m - l:
                        # add a tuple giving the lower and upper bounds of the
                        # indices of the subtree's nodes at this level
                        temp.append((left, right + 1))
                        self.log.debug(
                            msg="added %d indices at depth %d" % (left - right + 1, d)
                        )
                    # get the lower and upper bounds at the next level
                    left *= 2
                    right = 2 * right + 1
                self.log.debug(msg="indexing %s^%d-%d complete" % (prefix, m, n))
            # sort temp list by lower bound values
            temp.sort(key=itemgetter(0))
            # get reference to the correct set
            s = self.sets(af)
            # loop through the resulting range entries and
            # merge into the smallest set of entries
            lower, upper = None, None
            for next_lower, next_upper in temp:
                if lower is None:
                    lower = next_lower
                    upper = next_upper
                    continue
                if upper < next_lower:
                    s.add((lower, upper))
                    lower = next_lower
                upper = next_upper
            else:
                if upper is not None:
                    s.add((lower, upper))
        self.log_init_done()

    def __contains__(self, item):
        if isinstance(item, tuple):
            version = item[0]
            index = item[1]
        else:
            try:
                prefix, index = self.index_of(prefix=item)
            except ValueError as e:
                self.log.error(msg=str(e))
                raise e
            version = prefix.version
        for lower, upper in self.sets(version):
            if lower <= index < upper:
                return True
        return False

    def __iter__(self):
        for version in (4, 6):
            for lower, upper in self.sets(version):
                index = lower
                while index < upper:
                    yield (version, index)
                    index += 1

    def __len__(self):
        count = 0
        for version in (4, 6):
            for lower, upper in self.sets(version):
                count += upper - lower
        return count

    def _iter_ranges(self, versions=(4, 6), length=None):
        if isinstance(versions, int):
            versions = (versions,)
        for version in versions:
            for lower, upper in self.sets(version):
                if length is None:
                    yield (version, lower, upper)
                else:
                    if self.length_from_index(lower) == length:
                        yield (version, lower, upper)

    @classmethod
    def _from_iterable(cls, it):
        self = cls({})
        for item in it:
            af = "ipv%d" % item[0]
            lower = item[1]
            try:
                upper = item[2]
            except IndexError:
                upper = lower + 1
            try:
                self.sets(af).add((lower, upper))
            except KeyError as e:
                self.log.error(msg=str(e))
                raise e
        self._merge()
        return self

    def _merge(self):
        for version in (4, 6):
            s = self.sets(version)
            merged = set()
            try:
                temp = sorted(list(s), key=itemgetter(0))
            except IndexError:
                continue
            lower, upper = None, None
            for next_lower, next_upper in temp:
                if lower is None:
                    lower = next_lower
                    upper = next_upper
                    continue
                if upper < next_lower:
                    merged.add((lower, upper))
                    lower = next_lower
                upper = next_upper
            else:
                if upper is not None:
                    merged.add((lower, upper))
            s = merged
        return

    def __and__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        it = list()
        for i_version, i_lower, i_upper in other._iter_ranges():
            for j_version, j_lower, j_upper in self._iter_ranges():
                if i_version == j_version:
                    v = i_version
                    l = max(i_lower, j_lower)
                    u = min(i_upper, j_upper)
                    if l <= u:
                        it.append((v, l, u))
        return self._from_iterable(it)

    def __or__(self, other):
        if not isinstance(other, self.__class__):
            return NotImplemented
        it = list(self._iter_ranges()) + list(other._iter_ranges())
        return self._from_iterable(it)

    def iter_add(self, it):
        for item in it:
            prefix = self.check_val(item)
            af = "ipv%d" % prefix.version
            lower = item[1]
            try:
                upper = item[2]
            except IndexError:
                upper = lower + 1
            try:
                self.sets(af).add((lower, upper))
            except KeyError as e:
                self.log.error(msg=str(e))
                raise e
        return self

    def sets(self, af=None):
        try:
            return self._sets[af]
        except KeyError as e:
            try:
                return self._sets["ipv%d" % af]
            except KeyError as f:
                self.log.warning(msg=str(e))
                self.log.error(msg=str(f))
                raise f

    def prefixes(self):
        for version, index in self:
            yield self.indexed_by(index=index, af="ipv%d" % version)

    def data(self, aggregate=True):
        data = {"ipv4": [], "ipv6": []}
        if aggregate:
            # build a list of subtrees
            subtrees = list()
            # iterate through the ranges in the set
            for version, lower, upper in self._iter_ranges():
                root = lower
                # iterate through the root nodes in each range
                while root < upper:
                    # get the length of the root node
                    length = self.length_from_index(root)
                    left = right = root
                    depth = 0
                    while True:
                        found = False
                        # get the subtree min and max at the next depth
                        left *= 2
                        right = 2 * right + 1
                        # get the ranges at the current depth
                        for v, next_lower, next_upper in self._iter_ranges(
                            versions=version, length=length + depth + 1
                        ):
                            if next_lower <= left and next_upper >= right:
                                found = True
                                break
                        if found:
                            depth += 1
                        else:
                            break
                    # check whether the subtree is already covered
                    covered = False
                    for st in subtrees:
                        # check vertically
                        if (st["min"] <= length) and (length + depth <= st["max"]):
                            # check horizontally
                            left = st["root"] * 2 ** (length - st["min"])
                            right = (st["root"] + 1) * 2 ** (length - st["min"]) - 1
                            if left <= root <= right:
                                covered = True
                    if not covered:
                        subtrees.append(
                            {
                                "version": version,
                                "root": root,
                                "min": length,
                                "max": length + depth,
                            }
                        )
                    root += 1
            # loop through the list of subtrees, and aggregate
            while subtrees:
                st = subtrees.pop()
                branch = st["root"] % 2
                for st_other in subtrees:
                    if not st["version"] == st_other["version"]:
                        continue
                    if not st["min"] == st_other["min"]:
                        continue
                    if not st["max"] == st_other["max"]:
                        continue
                    diff = st["root"] - st_other["root"]
                    if (branch == 0 and diff == -1) or (branch == 1 and diff == 1):
                        subtrees.remove(st_other)
                        st["root"] = (st["root"] - branch) // 2
                        subtrees.append(st)
                        break
                else:
                    af = "ipv%d" % st["version"]
                    prefix = self.indexed_by(st["root"], af=af)
                    length = prefix.prefixlen
                    entry = {"prefix": str(prefix)}
                    if st["min"] != length:
                        entry["greater-equal"] = st["min"]
                    if st["max"] != length:
                        entry["less-equal"] = st["max"]
                    data[af].append(entry)
        else:
            for prefix in self.prefixes():
                af = "ipv%d" % prefix.version
                prefix = "%s/%d" % (prefix.network_address, prefix.prefixlen)
                data[af].append({"prefix": prefix})
        return data

    def meta(self, key=None, strict=False):
        if key:
            try:
                return self._meta[key]
            except KeyError as e:
                if strict:
                    self.log.error(msg=str(e))
                    raise e
                else:
                    self.log.debug(msg=str(e))
                    return None
        else:
            return self._meta

    @staticmethod
    def index_of(prefix=None):
        prefix = ipaddress.ip_network(prefix)
        p = int(prefix.network_address)
        l = prefix.prefixlen
        h = prefix.max_prefixlen
        index = 2 ** l + p // 2 ** (h - l)
        return prefix, index

    @staticmethod
    def indexed_by(index=None, af=None):
        address_families = {
            "ipv4": {"size": 32, "cls": ipaddress.IPv4Network},
            "ipv6": {"size": 128, "cls": ipaddress.IPv6Network},
        }
        assert isinstance(index, int)
        assert af in address_families
        h = int(address_families[af]["size"])
        cls = address_families[af]["cls"]
        l = index.bit_length() - 1
        p = index * 2 ** (h - l) - 2 ** h
        return cls((p, l))

    @staticmethod
    def length_from_index(index):
        return index.bit_length() - 1

    @staticmethod
    def parse_prefix_range(expr=None):
        pattern = (
            r"^(?P<prefix>((\d+(\.\d+){3})|([0-9a-fA-F:]{2,40}))/\d+)"
            r"(\^(?P<ge>\d+)-(?P<le>\d+))?$"
        )
        match = re.match(pattern, expr)
        if match:
            prefix = ipaddress.ip_network(str(match.group("prefix")))
            afi = "ipv%d" % prefix.version
            entry = {"prefix": prefix}
            try:
                entry["greater-equal"] = int(match.group("ge"))
                entry["less-equal"] = int(match.group("le"))
            except (IndexError, TypeError):
                pass
        else:
            raise ValueError("no match found")
        return {afi: [entry]}



import inspect


class IRRBase:
    """
    Base class for performing IRR queries
    """

    def __enter__(self):
        return self

    def __exit__(self, typ, value, traceback):
        pass

    def set_sources(self, *sources):
        """set sources to the specified list"""
        raise NotImplementedError(
            "{} does not implement {}".format(
                self.__class__.__name__, inspect.currentframe().f_code.co_name
            )
        )

    def iter_sets(self, objs, expand=True):
        """
        Return members of an as-set or route-set.
        if expand is true, also recursively expand members of all sets within the named set.
        """
        raise NotImplementedError(
            "{} does not implement {}".format(
                self.__class__.__name__, inspect.currentframe().f_code.co_name
            )
        )

    def get_sets(self, objs, expand=True):
        """
        Return members of an as-set or route-set.
        if expand is true, also recursively expand members of all sets within the named set.
        """
        return list(self.iter_sets(objs, expand))

    def iter_routes(self, objs, proto=4):
        """get routes for specified object"""
        raise NotImplementedError(
            "{} does not implement {}".format(
                self.__class__.__name__, inspect.currentframe().f_code.co_name
            )
        )

    def get_routes(self, objs, proto=4):
        """get routes for specified object"""
        return list(self.iter_routes(objs, proto))

    def iter_prefixes(self, as_sets, proto=4):
        """
        get prefix list for specified as-set(s)
        """
        raise NotImplementedError(
            "{} does not implement {}".format(
                self.__class__.__name__, inspect.currentframe().f_code.co_name
            )
        )

    def get_prefixes(self, as_sets, proto=4):
        """
        get prefix list for specified as-set(s)
        """
        return list(self.iter_prefixes(as_sets, proto))



# Copyright (C) 2016 Matt Griswold <grizz@20c.com>
#
# This file is part of bgpfu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import collections.abc
import ipaddress

from bgpfu.prefixlist import PrefixListBase


def _try_combine(aggregate, current):
    "try combining and replacing the last element on the aggregate list"
    if aggregate and aggregate[-1]:
        supernet = aggregate[-1].supernet()
        if supernet == current.supernet():
            aggregate[-1] = supernet
            return True
    return False


def _do_aggregate(prefixlist):
    if len(prefixlist) <= 1:
        return prefixlist
    prefixlist = sorted(prefixlist)

    # TODO check for default and skip it?

    aggregate = []
    while True:
        current = None
        for pfx in prefixlist:
            if not current:
                current = pfx
                continue

            if current.overlaps(pfx):
                continue

            # try joining 2
            supernet = current.supernet()
            if supernet == pfx.supernet():
                current = supernet
                continue

            # nothing to combine, shift
            aggregate.append(current)
            current = pfx

        if current:
            if not _try_combine(aggregate, current):
                aggregate.append(current)

        if len(aggregate) == len(prefixlist):
            return aggregate

        prefixlist = aggregate
        aggregate = []


class SimplePrefixList(PrefixListBase, collections.abc.MutableSequence):
    """
    Simple PrefixList implemenatation using collections
    *NOTE* loses prefix length info on aggregate
    """

    def __init__(self, prefixes=None):
        if prefixes:
            self._prefixes = list(map(ipaddress.ip_network, list(map(str, prefixes))))
        else:
            self._prefixes = []

    def __getitem__(self, i):
        return self._prefixes[i]

    def __setitem__(self, i, v):
        self._prefixes[i] = self.check_val(v)

    def insert(self, i, v):
        self._prefixes.insert(i, self.check_val(v))

    def iter_add(self, it):
        for v in it:
            self._prefixes.append(self.check_val(v))

    def __delitem__(self, i):
        del self._prefixes[i]

    def __len__(self):
        return len(self._prefixes)

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            return self._prefixes == other._prefixes
        raise TypeError("object not PrefixList type")

    def __ne__(self, other):
        return not self == other

    def __str__(self):
        return str(self._prefixes)

    @property
    def ipv4(self):
        return [p for p in self._prefixes if p.version == 4]

    @property
    def ipv6(self):
        return [p for p in self._prefixes if p.version == 6]

    def str_list(self):
        return list(map(str, self._prefixes))

    def aggregate(self):
        "returns a PrefixList containing the result of aggregating the list"

        if len(self._prefixes) == 1:
            return self.__class__(self._prefixes)

        # v4 = sorted(self._prefixes)
        v4 = [p for p in self._prefixes if p.version == 4]
        v6 = [p for p in self._prefixes if p.version == 6]

        v4 = _do_aggregate(v4)
        v6 = _do_aggregate(v6)
        return self.__class__(v4 + v6)



# Copyright (C) 2016 Matt Griswold <grizz@20c.com>
#
# This file is part of bgpfu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import inspect
import re

import pkg_resources


class Output:
    """
    abstract writing to different output formats
    """

    @property
    def available_formats(self):
        re_out = re.compile(r"^output_(?P<type>\w+)$")
        fmts = []

        for name, obj in inspect.getmembers(type(self), inspect.isfunction):
            match = re_out.match(name)
            if match:
                fmts.append(match.group("type"))

        return fmts

    def write(self, fmt, fobj, data):
        """
        write data in specified format to fobj
        """
        func = "output_" + fmt
        if not hasattr(self, func):
            raise ValueError("unknown output format '%s'" % fmt)
        return getattr(self, func)(fobj, data)

    def load_file(self, filename):
        """
        load file from package
        """
        return pkg_resources.resource_string("bgpfu", filename)

    def _init_jinja(self, tmpl_name):
        tmpl = self.load_file(tmpl_name + ".j2")

    def output_juniper(self, jobj, data):
        eng = self._init_jinja("juniper")

    def output_txt(self, fobj, data):
        if isinstance(data, list):
            fobj.write("\n".join(data))
            fobj.write("\n")
            return



# Copyright (C) 2016 Matt Griswold <grizz@20c.com>
#
# This file is part of bgpfu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import logging
import re

import gevent
from pkg_resources import get_distribution

from bgpfu.io import Empty, Queue, select, socket
from bgpfu.irr import IRRBase
from bgpfu.prefixlist import SimplePrefixList as PrefixList


class IRRClient(IRRBase):
    """
    IRR client, uses pipelining
        supports keepalive
    """

    def __init__(self):
        self.keepalive = True
        self.host = "whois.radb.net"
        self.host = "rr.ntt.net"
        self.port = 43

        self.re_res = re.compile(r"(?P<state>[ACDEF])(?P<len>\d*)(?P<msg>[\w\s]*)$")

        self.log = logging.getLogger(__name__)

        self.sckt = None
        self._send_queue = Queue()
        self._send_thread = None
        self._req_sent = 0

    def __enter__(self):
        self.connect()
        return self

    def __exit__(self, typ, value, traceback):
        self.sckt.shutdown(socket.SHUT_RDWR)
        self.sckt.close()
        self._send_thread.kill()
        self._send_thread.get()

    def connect(self):
        self.sckt = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.sckt.connect((self.host, self.port))
        self.sckt.setblocking(False)

        if self.keepalive:
            self.sckt.send(b"!!\n")

        self._send_thread = gevent.spawn(self._process_send_queue)

        self.query_one("!nBGPFU-{}".format(get_distribution("bgpfu").version))

    def make_set_query(self, obj, expand=True):
        """
        if expand is true, also recursively expand members of all sets within the named set.
        """
        q = "!i" + obj
        if expand:
            q = q + ",1"
        return q

    def make_route_query(self, obj, proto=4):
        proto = int(proto)
        if proto == 4:
            return "!g" + obj
        elif proto == 6:
            return "!6" + obj

        raise ValueError("unknown protocol '%s'" % str(proto))

    def parse_response(self, response):
        self.log.debug("response %s", response)
        match = self.re_res.match(response)
        if not match:
            raise RuntimeError(f"invalid response '{response}'")

        state = match.group("state")
        if state == "A":
            return int(match.group("len"))
        elif state == "C":
            return False
        elif state == "D":
            self.log.warning("skipping key not found")
            return False
        elif state == "E":
            raise KeyError("multiple copies of key in database")
        elif state == "F":
            if match.group("msg"):
                msg = match.group("msg").strip()
            else:
                msg = "unknown error"
            raise RuntimeError(msg)

        raise RuntimeError(f"invalid response '{response}'")

    def iter_query(self, querylist):
        """
        performs a query, returns a generator
        """
        if not self.sckt:
            raise OSError("not connected")

        self.log.debug(f"QUERY {querylist}")
        self._queue_query(querylist)
        gevent.sleep(0.0001)
        yield from self._pipeline_read()

    def query_one(self, query):
        """
        performs a query, checks for exactly 1 result, returns it
        """
        res = list(self.iter_query((query,)))

        if len(res) > 1:
            raise ValueError(f"query_one returned {len(res)} results")

        elif len(res) == 1:
            return res[0]

        return None

    def query(self, querylist):
        """
        performs a query, returns a list
        """
        if isinstance(querylist, str):
            querylist = (querylist,)

        return list(self.iter_query(querylist))

    def _queue_query(self, querylist):
        if isinstance(querylist, str):
            querylist = (querylist,)
        # throw if queue is full
        self._send_queue.put_nowait(querylist)

    def _get_send_queue(self):
        max_chunk = 10
        idx = 0

        # force list, could be tuples on queue
        querylist = []
        # block until we get at least one
        querylist.extend(self._send_queue.get(True))

        # try to add any additional requests
        while True:
            try:
                if idx >= max_chunk:
                    break
                querylist.extend(self._send_queue.get_nowait())
                idx += 1

            except Empty:
                break
        return querylist

    def _process_send_queue(self):
        while True:
            self.log.debug("blocking on send queue")
            querylist = self._get_send_queue()

            qlen = len(querylist)
            self._req_sent += qlen
            # self.log.debug("sending {}/{}".format(qlen, self._req_sent))
            q = "\n".join(querylist) + "\n"
            ttl = 0
            sz = len(q)
            while ttl < sz:
                sent = self.sckt.send(q[ttl:].encode("ascii"))
                if not sent:
                    raise RuntimeError("socket connection broken")
                ttl = ttl + sent

            self.log.debug(f"sent '{q.rstrip()}' ]{ttl}]")

    def _pipeline_read(self):
        if not self._req_sent:
            raise RuntimeError("read called without a queued request")

        reqno = 0
        buf = ""
        while reqno < self._req_sent:
            self.log.debug(f"processing {reqno} of {self._req_sent}")
            res, buf = self._read_res(buf)
            reqno += 1
            if not res:
                if self._req_sent == 1:
                    break
                continue
            yield res

        self._req_sent = 0

    def _read_res(self, buf=""):
        """
        read next response
        """
        chunks = []
        ttl = 0
        chunk_size = 4096

        # if we already have data, don't wait at all on the read
        if buf:
            timeout = 0
        else:
            timeout = 10

        chunk = ""
        readable = select.select([self.sckt], [], [], timeout)[0]

        if readable:
            chunk = self.sckt.recv(chunk_size).decode("utf-8")

        #        self.log.debug("-- BUF -----\n{}------------".format(buf))
        #        self.log.debug("-- CHUNK ---\n{}------------".format(chunk))
        if buf:
            chunk = buf + chunk

        idx = chunk.find("\n")
        if idx == -1:
            raise RuntimeError("no data to read, but no response in buffer")

        response = chunk[:idx]
        chunk = chunk[idx + 1 :]

        sz = self.parse_response(response)
        if not sz:
            return None, chunk

        ttl = len(chunk)
        chunks.append(chunk)

        while ttl <= sz:
            #            self.log.debug("ttl %d, sz %d", ttl, sz)
            readable = select.select([self.sckt], [], [])[0]
            if not readable:
                raise RuntimeError("socket connection broken")

            chunk = self.sckt.recv(chunk_size).decode("utf-8")

            if chunk == "":
                raise RuntimeError("socket connection broken")
            chunks.append(chunk)
            ttl = ttl + len(chunk)

        # slice off remaining buffer
        buf_sz = ttl - sz
        buf = chunks[-1][-buf_sz:]
        chunks[-1] = chunks[-1][:-buf_sz]

        # try to strip off command complete
        # FIXME - need to stop eating the last C on a result, it screws up the results
        buf = buf.lstrip("C")
        buf = buf.lstrip("\n")

        # self.log.debug("-- FINRES ---\n{}\n-- /FINRES ----".format(''.join(chunks)))
        # self.log.debug("-- FINBUF -----\n{}-- /FINBUF ----".format(buf))
        # self.log.debug("ttl %d, sz %d", ttl, sz)
        # self.log.debug("remaining %d '%s'", ttl - sz, buf)
        return "".join(chunks), buf

    def set_sources(self, *sources):
        """set sources to the specified list"""
        res = self.query("!s{}".format(",".join(sources)))
        # FIXME - stop eating the last C on a command
        # if len(res) != 1:
        #     raise RuntimeError("source query returned multiple results")

        # return res[0]
        return None

    def iter_sets(self, objs, expand=True):
        """
        Return members of an as-set or route-set.
        if expand is true, also recursively expand members of all sets within the named set.
        """
        if isinstance(objs, str):
            objs = (objs,)

        sets = []
        querylist = []

        for each in objs:
            q = "!i" + each
            if expand:
                q = q + ",1"
            self.log.debug(f"ADDING {q}")
            querylist.append(q)

        self.log.debug(f"QUERLLLLLLL {querylist}")
        for res in self.iter_query(querylist):
            sets += res.split()

        return set(sets)

    def iter_routes(self, obj, proto=4):
        """get routes for specified object"""
        proto = int(proto)
        if proto == 4:
            q = "!g"
        elif proto == 6:
            q = "!6"
        else:
            raise ValueError("unknown protocol '%s'" % str(proto))

        q += obj
        prefixes = []
        for res in self.iter_query((q,)):
            if res:
                yield res.split()

    def iter_prefixes(self, as_sets, proto=4):
        """get prefix list for specified as-set(s)"""
        if isinstance(as_sets, str):
            as_sets = (as_sets,)
        querylist = list(map(self.make_set_query, as_sets))

        # get routes for each AS SET, put directly onto send queue
        for res in self.iter_query(querylist):
            self._queue_query(list(map(self.make_route_query, res.split())))

        # FIXME - need better handoff to io thread
        gevent.sleep(0.0001)

        # read results from all requests
        for res in self._pipeline_read():
            if res:
                yield res.split()



import inspect
import ipaddress

from bgpfu.base import BaseObject


class PrefixListBase(BaseObject):
    def __init__(self, prefix_ctor=ipaddress.ip_network, prefixes=None, aggregate=True):
        super().__init__()
        self.prefix_ctor = prefix_ctor
        self.aggregate = aggregate

    def check_val(self, v):
        """check value, call ctor if needed"""
        if not isinstance(v, (ipaddress.IPv4Network, ipaddress.IPv6Network)):
            return ipaddress.ip_network(str(v))
        return v

    def make_prefix(self, prefix):
        return self.prefix_ctor(prefix)

    def add(self, prefix, meta={}):
        raise NotImplementedError(
            "{} does not implement {}".format(
                self.__class__.__name__, inspect.currentframe().f_code.co_name
            )
        )

    def iter_add(self, prefix, meta={}):
        raise NotImplementedError(
            "{} does not implement {}".format(
                self.__class__.__name__, inspect.currentframe().f_code.co_name
            )
        )

    def str_list(self):
        raise NotImplementedError(
            "{} does not implement {}".format(
                self.__class__.__name__, inspect.currentframe().f_code.co_name
            )
        )


# TODO add get functions



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['bgpfu', 'bgpfu.irr', 'bgpfu.prefixlist']

package_data = \
{'': ['*'], 'bgpfu': ['templates/*']}

install_requires = \
['Jinja2>=2.11.3,<3.0.0',
 'click>=7.1.2,<8.0.0',
 'gevent>=21.1.2,<22.0.0',
 'munge>=1.1.0,<2.0.0',
 'py-radix>=0.10.0,<0.11.0']

entry_points = \
{'console_scripts': ['bgpfu = bgpfu.cli:cli']}

setup_kwargs = {
    'name': 'bgpfu',
    'version': '0.3.0',
    'description': 'BGP toolkit',
    'long_description': '\n[![PyPI](https://img.shields.io/pypi/v/bgpfu.svg?maxAge=60)](https://pypi.python.org/pypi/bgpfu)\n[![PyPI](https://img.shields.io/pypi/pyversions/bgpfu.svg?maxAge=600)](https://pypi.python.org/pypi/bgpfu)\n[![Tests](https://github.com/bgpfu/bgpfu/workflows/tests/badge.svg)](https://github.com/bgpfu/bgpfu)\n[![Codecov](https://img.shields.io/codecov/c/github/bgpfu/bgpfu/master.svg?maxAge=60)](https://codecov.io/github/bgpfu/bgpfu)\n\n\n# BGP FU\n\n![BGPFU](/../gh-pages/images/BGP-FU-Logo-RGB-resized.png?raw=true)\n\nBGP FU is a toolbelt to assist with the automatic creation of safe prefix-filters.\n\nBGP FU can ingest data from various sources such as IRRs/RPKI/other, transpose those\naccording to your policy preferences, and generate partial router configurations.\n\nThe jinja2 templating language is used to describe vendor specific output.\n\n## Contact\n\nJoin us on `irc.terahertz.net` in `#BGPFU`\n',
    'author': 'Matt Griswold',
    'author_email': 'grizz@20c.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/bgpfu/bgpfu/',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6.2,<4.0.0',
}


setup(**setup_kwargs)



# Copyright (C) 2016 Matt Griswold <grizz@20c.com>
#
# This file is part of bgpfu
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import ipaddress
import logging

import click

import bgpfu.client
from bgpfu.output import Output
from bgpfu.prefixlist import SimplePrefixList as PrefixList
from bgpfu.prefixlist.set import PrefixSet

from .roa import RoaTree

# global output formatter
outfmt = Output()


def common_options(f):
    f = click.option("--debug", is_flag=True, default=False)(f)
    return f


def as_set_options(f):
    f = click.option(
        "--skip-as",
        help="comma separated list of ASNs to not include in results",
        default=False,
    )(f)
    return f


def output_options(f):
    f = click.option(
        "--output-format",
        "-f",
        help="output format",
        type=click.Choice(outfmt.available_formats),
        default="txt",
        show_default=True,
    )(f)
    f = click.option("--output", "-o", help="output file", default="-")(f)
    return f


@click.group()
# @common_options
@click.version_option()
@click.pass_context
def cli(ctx):
    pass


@cli.command()
@click.pass_context
# @connect_options
@common_options
@click.argument("prefix", nargs=1)
@click.argument("asn", nargs=1)
@click.option("--rib-file", help="use bgp rib file")
@click.option("--rpki-file", help="use rpki json file")
def check_roa(ctx, prefix, asn, rib_file, rpki_file, **kwargs):
    """check roa for prefix"""
    prefix = ipaddress.ip_network(prefix)
    asn = int(asn)

    if rib_file and rpki_file:
        raise ValueError("bgp-rib and rpki-json are mutually exclusive")

    if rib_file:
        db = RoaTree(rib_file=rib_file)

    if rpki_file:
        db = RoaTree(rpki_file=rpki_file)

    print(f"check_roa {prefix} {asn}")
    rv = db.validation_state(prefix, asn)
    print(rv)


@cli.command()
@click.pass_context
# @connect_options
@common_options
@click.argument("query", nargs=1)
def raw(ctx, query, **kwargs):
    """raw irr query"""
    if kwargs.get("debug", False):
        logging.basicConfig(level=logging.DEBUG)

    bgpfuc = bgpfu.client.IRRClient()
    with bgpfuc as c:
        data = c.query(query)
        print(data)


@cli.command()
@common_options
@as_set_options
@click.argument("as-set", nargs=-1)
@click.pass_context
def as_set(ctx, as_set, **kwargs):
    """expand an as-set"""
    if kwargs.get("debug", False):
        logging.basicConfig(level=logging.DEBUG)
    bgpfuc = bgpfu.client.IRRClient()
    with bgpfuc as c:
        sets = c.get_sets(as_set)

    print("\n".join(sets))


@cli.command()
# @connect_options
@common_options
@output_options
@click.option(
    "--aggregate/--no-aggregate",
    "-A",
    help="aggregate prefixes",
    is_flag=True,
    default=True,
)
@click.option("--sources", help="use only specified sources (default is all)")
@click.option("--4", "-4", "proto", help="use IPv4", flag_value="4", default=True)
@click.option("--6", "-6", "proto", help="use IPv6", flag_value="6")
@click.option("--fancy", help="use fancy set", flag_value=True)
@click.argument("as-set", nargs=-1)
@click.pass_context
def prefixlist(ctx, as_set, output, output_format, proto, **kwargs):
    """get prefix list for specified as-sets"""
    if kwargs.get("debug", False):
        logging.basicConfig(level=logging.DEBUG)

    if kwargs.get("fancy", 0):
        prefixes = PrefixSet(aggregate=kwargs["aggregate"])
    else:
        prefixes = PrefixList()
    bgpfuc = bgpfu.client.IRRClient()
    with bgpfuc as c:
        if kwargs.get("sources", False):
            c.set_sources(kwargs["sources"])
        #        prefixes = c.get_prefixes(as_set, proto)

        for chunk in c.iter_prefixes(as_set, proto):
            prefixes.iter_add(chunk)

        if not kwargs["fancy"] and kwargs["aggregate"]:
            prefixes = prefixes.aggregate()

        with click.open_file(output, "w") as fobj:
            outfmt.write(output_format, fobj, prefixes.str_list())

        print(f"LEN {len(prefixes)}")



import ipaddress
import json
import logging

import radix

logger = logging.getLogger(__name__)


def parse_roa(roa):
    """Parses an roa record in place and returns the result."""
    try:
        roa["prefix"] = ipaddress.ip_network(roa["prefix"])
        if not isinstance(roa["asn"], int):
            roa["asn"] = int(roa["asn"].replace("AS", ""))
        roa["maxLength"] = int(roa["maxLength"])
        # check bounds on asn
        if not 0 <= roa["asn"] < 2 ** 32:
            raise ValueError(f"asn {roa['asn']} is out of range")
    except ValueError as exc:
        logger.warning(str(exc))
        return None

    return roa


def parse_rib_line(line):
    """
    Parses a line from a rib file and yields prefix_str, roa.
    """

    if line[0] == "#":
        return
    try:
        prefix_str, asns = tuple(line.split())
        prefix = ipaddress.ip_network(prefix_str)
    except Exception as exc:
        line = line.rstrip()
        logger.warning(f"Could not parse line '{line}': {exc}")
        return

    for asn in asns.split("|"):
        try:
            yield prefix_str, dict(
                asn=int(asn), prefix=prefix, maxLength=prefix.prefixlen
            )
        except (TypeError, ValueError):
            line = line.rstrip()
            logger.warning(f"Could not parse asn from line '{line}'")


def _add_node(tree, prefix_str, roa):
    """Add a node to a radix tree under prefix_str."""
    node = tree.search_exact(prefix_str)
    if not node:
        node = tree.add(prefix_str)
    node.data.setdefault("roas", []).append(roa)


class RoaTree:
    def __init__(self, rib_file=None, rpki_file=None):
        """Creates an instance of RoaTree from passed file."""
        self.meta = dict()

        if rib_file and rpki_file:
            raise ValueError("rib and rpki are mutually exclusive")

        if rib_file:
            self.load_rib_file(rib_file)

        if rpki_file:
            self.load_rpki_file(rpki_file)

    def load_rib_file(self, filename, merge=False):
        """Loads a rib file into the tree.

        Merge=True adds to the current tree
        """
        if merge:
            tree = self._tree
        else:
            tree = radix.Radix()

        with open(filename) as fh:
            for line in fh.readlines():
                for prefix_str, roa in parse_rib_line(line):
                    _add_node(tree, prefix_str, roa)

        if merge:
            self.meta.setdefault("filename", []).append(filename)
        else:
            self.meta["filename"] = [filename]
        self.meta["type"] = "rib"
        self._tree = tree

    def load_rpki_file(self, filename):
        """Loads a rpki json file."""
        with open(filename) as fh:
            data = json.load(fh)
        tree = radix.Radix()

        for roa in data["roas"]:
            prefix_str = roa["prefix"]
            roa = parse_roa(roa)
            if not roa:
                continue
            _add_node(tree, prefix_str, roa)

        self.meta = data["metadata"]
        self.meta["filename"] = [filename]
        self.meta["type"] = "rib"
        self._tree = tree

    def check_invalid(self, *args, **kwargs):
        """Check if invalid.

        Returns True for invalid, False if valid or not found.
        """
        return self.validation_state(*args, **kwargs)["state"] == "invalid"

    def validate_best(self, prefix, origin):
        """Validate prefix and orgin against most specific match only."""
        covered_roas = []
        prefix = ipaddress.ip_network(prefix)
        prefix_str = str(prefix)
        vrp = self._tree.search_best(prefix_str)
        if not vrp:
            return {"state": "notfound"}

        vrp_prefix = ipaddress.ip_network(vrp.prefix)
        if not prefix.subnet_of(vrp_prefix):
            return {"state": "notfound"}

        for roa in vrp.data["roas"]:
            if origin == roa["asn"]:
                return {"state": "valid", "roa": roa}

            covered_roas.append(roa)

        return {"state": "invalid", "roas": covered_roas}

    def validation_state(self, prefix, origin, check_maxlength=True):
        """
        prefix is the to-be-tested prefix
        origin is the origin asn to be used in the test
        """

        tree = self._tree
        prefix = ipaddress.ip_network(prefix)
        prefix_str = str(prefix)

        best = tree.search_best(prefix_str)
        if not tree.search_best(prefix_str):
            return {"state": "notfound"}

        covered_roas = []
        worst_prefix = tree.search_worst(prefix_str).prefix

        for vrp in tree.search_covered(worst_prefix):
            vrp_prefix = ipaddress.ip_network(vrp.prefix)

            if not prefix.subnet_of(vrp_prefix):
                continue

            for roa in vrp.data["roas"]:
                covered_roas.append(roa)
                if origin != roa["asn"]:
                    continue

                if vrp.prefixlen > prefix.prefixlen:
                    continue

                if check_maxlength and prefix.prefixlen > roa["maxLength"]:
                    continue
                return {"state": "valid", "roa": roa}

        return {"state": "invalid", "roas": covered_roas}
